
let init () =
	(* Clear color and depth buffer *)
	GlClear.color (0.0, 0.0, 0.0);
	GlClear.clear [`color;`depth];
	GlDraw.shade_model `smooth;
	List.iter Gl.enable [`depth_test; `cull_face; `light0; `texture_gen_s; `texture_gen_t];
	GlTex.gen `s (`mode `sphere_map);
	GlTex.gen `t (`mode `sphere_map);
	(* Get and select a texture object ID *)
	let texid = GlTex.gen_texture () in
		GlTex.bind_texture `texture_2d texid;
	(* Load texture into string from file, create a GlPix image, create a "raw" alias to this image,
	   and blit the contents of the texture string into the raw representaton of the GlPix image *)
	let w,h,s = Loadraw.loadRawImg "./data/objectline/base.jpeg.raw" in
	let img = GlPix.create `ubyte ~height:h ~width:w ~format:`rgba in
	let raw = GlPix.to_raw img in
		Raw.sets_string raw ~pos:0 s;
		(* Use trilinear interpolation (GL_LINEAR_MIPMAP_LINEAR)*)
		List.iter (GlTex.parameter ~target:`texture_2d)  [`min_filter `linear_mipmap_linear; `mag_filter `linear];
		GluMisc.build_2d_mipmaps ~internal:4 img;
	(* Load model and use list of vertices to create display list *)
	let model = Loadraw.loadRaw3DS "./data/objectline/model.3ds.raw" in
	let num_face = (Array.length model) / 8 in
	let list = GlList.create `compile in
		GlDraw.begins `triangles;
		for i = 0 to num_face - 1 do
			let j = i*8 in
			let v = (model.(j), model.(j + 1), model.(j + 2))
			and n = (model.(j + 3), model.(j + 4),model.(j + 5)) in
			GlDraw.normal3 n;
			GlDraw.vertex3 v;
		done;
		GlDraw.ends ();
	GlList.ends ();
	list
;;


let getFps starttime  fps counter =
	if counter = 100 then 
		let endtime = Glfw.get_time () in
		let newfps = (float counter)  /. (endtime -. starttime) in 	
		(endtime, (newfps +. fps) /. 2.0, 0)
	else
		(starttime, fps, counter + 1)
;;

let pan fps view angle =
	let ifps = 1.0 /. fps in
	let nview = view +. ifps in
	let camera = [| 0.0; 120.0; 240.0 *. (cos (view /. 3.0))|] in
	let dir = [|0.0; 0.0; 0.0|]  
	and light = [|camera.(0); camera.(1); camera.(2); 1.0;|]
	and nangle = angle -. (ifps *. 360.0 *. 0.2) in
	(camera, dir, light, nview, nangle)
;;

let display list camera dir light angle =
	(* Clear color and depth buffer *)
	GlClear.color (0.0, 0.0, 0.0);
	GlClear.clear [`color;`depth];
    (* Select and setup the projection matrix *)
    GlMat.mode `projection;
    GlMat.load_identity ();
    GluMat.perspective ~fovy:(45.0) ~aspect:(4.0 /. 3.0) ~z:(1.0 ,1000.0);
    (* Select and setup the modelview matrix *)
    GlMat.mode `modelview;
	GlMat.load_identity();
    GluMat.look_at ~eye:(camera.(0),camera.(1),camera.(2))  ~center:(dir.(0), dir.(1), dir.(2)) ~up:(0.0, 0.0, 0.1);
	(* Set light *)
	GlLight.light ~num:0 (`position (light.(0), light.(1), light.(2), light.(3)));
	(* Render *)
	GlMat.push ();
	GlMat.rotate ~angle:angle ~z:1.0 ();
	
	GlClear.stencil 0;
	GlClear.clear [`stencil];
	Gl.enable `stencil_test;
	GlFunc.stencil_func `always ~ref:1 ~mask:(-1);
	GlFunc.stencil_op ~fail:`keep ~zfail:`keep ~zpass:`replace ();
	
	List.iter Gl.enable [`lighting; `texture_2d];
	GlList.call list;
	List.iter Gl.disable [`lighting; `texture_2d];
	
	GlFunc.stencil_func `notequal ~ref:1 ~mask:(-1);
	GlFunc.stencil_op ~fail:`keep ~zfail:`keep ~zpass:`replace ();
	
	GlDraw.line_width 8.0;
	GlDraw.polygon_mode ~face:`front `line;
	GlDraw.color (0.0, 1.0, 0.0);
	GlList.call list;
	GlDraw.color (1.0, 1.0, 1.0);
	GlDraw.polygon_mode ~face:`front `fill;
		
	List.iter Gl.disable [`stencil_test; `lighting; `texture_2d];
	GlMat.pop ();	
;;

let main () =
	try
	    (* Initialise GLFW *)
		if (Glfw.init () = 0) then failwith "glfwInit failed";
		(* Open OpenGL window *)
		if (Glfw.open_window 800 600 8 8 8 8 8 8 Glfw.window = 0) then begin
			Glfw.terminate ();
			failwith "glfwOpenWindow failed";	
		end;	
		(* Enable sticky keys *)
		Glfw.enable Glfw.sticky_keys;
		(* Disable vertical sync (on cards that support it) *)
		Glfw.swap_interval 0;
		(* Load texture and model, and get resulting display list *)
		let list = init () 
		(* initialize some variables we will need *)
		and ifps = 60.0
		and icounter = 0
		and istarttime = Glfw.get_time () 
		in
		(* rendering loop *) 
		let rec loop starttime fps counter view angle =
			let s, f, c = getFps starttime fps counter in 
			(* Set window caption *)
			let wtitle = Printf.sprintf "Objectline demo: %f fps (counter=%d)" f c in
			Glfw.set_window_title wtitle;
			(* pan function *)
			let camera, dir, light, nview, nangle = pan f view angle in
			display list camera dir light nangle;
		    (* Swap buffers -- also necessary for event polling *)					
		    Glfw.swap_buffers ();	
			(* Check if the ESC key was pressed or the window was closed *)
			if ((Glfw.get_key Glfw.key_esc) = 1)  or  ((Glfw.get_window_param Glfw.opened) = 0) then () else loop s f c nview nangle;		 
		in
		loop istarttime ifps icounter 0.0 0.0;
		(* Close OpenGL window and terminate GLFW *)
		Glfw.terminate ();
	with
	Loadraw.LoadRawImg_error s -> Printf.printf "Cannot load %s\n" s
	|Loadraw.LoadRaw3DS_error s -> Printf.printf "Cannot load %s\n" s
	|Failure s -> Printf.printf "Failure: %s\n" s
in
main ()
